#-----------------------------------------------------------------------------
# Copyright (c) 2015-2021, PyInstaller Development Team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
#
# The full license is in the file COPYING.txt, distributed with this software.
#
# SPDX-License-Identifier: Apache-2.0
#-----------------------------------------------------------------------------

import os
import sys


# enable Gio.AppInfo to find system default apps,
# i.e. for use with Gio.AppInfo.launch_default_for_uri
xdg_data_dirs_orig = os.environ.get("XDG_DATA_DIRS", None)
xdg_data_dirs_prepend = os.path.join(sys._MEIPASS, 'share')

if xdg_data_dirs_orig is not None:
    os.environ['XDG_DATA_DIRS_ORIG'] = xdg_data_dirs_orig
    os.environ['XDG_DATA_DIRS'] = xdg_data_dirs_prepend + os.pathsep + xdg_data_dirs_orig
else:
    os.environ['XDG_DATA_DIRS'] = xdg_data_dirs_prepend



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from gi.repository import Gtk


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/authentication_dialog.ui')
class AuthenticationDialog(Gtk.Window):
    __gtype_name__ = 'DtoolAuthenticationDialog'

    username_entry = Gtk.Template.Child()
    password_entry = Gtk.Template.Child()

    def __init__(self, apply=lambda username, password: None, username=None, password=None, **kwargs):
        super().__init__(**kwargs)
        if username:
            self.username_entry.set_text(username)
        if password:
            self.password_entry.set_text(password)
        self._apply = apply

    @Gtk.Template.Callback()
    def on_apply_clicked(self, widget):
        self._apply(self.username_entry.get_text(), self.password_entry.get_text())
        self.destroy()

    @Gtk.Template.Callback()
    def on_cancel_clicked(self, widget):
        self.destroy()



#
# Copyright 2023 Ashwin Vazhappilly
#           2022-2023 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import json
import logging
import types
from unittest import mock
from unittest.mock import patch, mock_open, MagicMock, ANY, AsyncMock
import pytest
import dtoolcore

from dtool_lookup_gui.models.settings import settings as app_settings


@pytest.mark.asyncio
async def test_app_id(running_app):
    assert running_app.get_application_id() == 'de.uni-freiburg.dtool-lookup-gui'


@pytest.fixture
def settings():
    """Provide test case with app settings."""
    yield app_settings


@pytest.mark.asyncio
async def test_do_toggle_logging(running_app):
    """Test do_toggle_logging action."""

    with patch('logging.disable') as mock_logging_disable:
        # Mock value to mimic the GVariant returned by the value parameter
        value = MagicMock()
        value.get_boolean.return_value = True  # Testing the True case

        # Mock action to mimic the Gio.SimpleAction
        action = MagicMock()

        # Perform the toggle logging action for enabling logging
        running_app.do_toggle_logging(action, value)

        # Assert that logging.disable was called with logging.NOTSET
        mock_logging_disable.assert_called_once_with(logging.NOTSET)
        action.set_state.assert_called_once_with(value)

        # Reset mocks to test the False case
        mock_logging_disable.reset_mock()
        action.set_state.reset_mock()
        value.get_boolean.return_value = False

        # Perform the toggle logging action for disabling logging
        running_app.do_toggle_logging(action, value)

        # Assert that logging.disable was called with logging.WARNING
        mock_logging_disable.assert_called_once_with(logging.WARNING)
        action.set_state.assert_called_once_with(value)


@pytest.mark.asyncio
async def test_do_set_loglevel(running_app):
    """Test do_set_loglevel action."""

    with patch.object(logging.getLogger(), 'setLevel') as mock_set_level:
        # Mock values for loglevel
        loglevel = 20  # Example log level (e.g., logging.INFO)

        # Mock value to mimic the GVariant returned by the value parameter
        value = MagicMock()
        value.get_uint16.return_value = loglevel

        # Mock action to mimic the Gio.SimpleAction
        action = MagicMock()
        action_state = MagicMock()
        action_state.get_uint16.return_value = 0
        action.get_state.return_value = action_state

        # Perform the set loglevel action
        running_app.do_set_loglevel(action, value)

        # Assert that the root logger's setLevel was called with the correct level
        mock_set_level.assert_called_once_with(loglevel)
        action.set_state.assert_called_once_with(value)


@pytest.mark.asyncio
async def test_do_set_logfile(running_app):
    """Test do_set_logfile action."""
    with patch('logging.FileHandler') as mock_file_handler, \
            patch.object(logging.getLogger(), 'addHandler') as mock_add_handler:
        # Mock values for logfile
        logfile = "/path/to/logfile.log"

        # Mock value to mimic the GVariant returned by the value parameter
        value = MagicMock()
        value.get_string.return_value = logfile

        # Mock action to mimic the Gio.SimpleAction
        action = MagicMock()
        action_state = MagicMock()
        action_state.get_string.return_value = ""
        action.get_state.return_value = action_state

        # Perform the set logfile action
        running_app.do_set_logfile(action, value)

        # Assert that logging.FileHandler was called with the correct path
        mock_file_handler.assert_called_once_with(logfile)

        # Assert that the FileHandler was added to the root logger
        mock_add_handler.assert_called_once_with(ANY)

        # Assert that the action's state was set to the new logfile value
        action.set_state.assert_called_once_with(value)


@pytest.mark.asyncio
async def test_do_reset_config(running_app, settings):
    """Test do_reset_config action."""
    with patch('os.remove') as mock_remove, \
            patch.object(running_app, 'emit') as mock_emit:  # Patch the emit method

        settings.reset = MagicMock()  # Mock the settings reset method

        # Perform the reset config action
        running_app.activate_action('reset-config')

        # The 'do_reset_config' action removes and replaces the dtool.json
        # config file.
        # Assert that os.remove has been called with the correct path
        mock_remove.assert_called_once_with(dtoolcore.utils.DEFAULT_CONFIG_PATH)

        # The 'do_reset_config' action calls the reset method on the
        # Gio.Setting property of the settings object.
        # Assert that the settings.reset was called once.
        settings.reset.assert_called_once()

        # The emit method is called when a signal is emitted.
        # The 'do_reset_config' action is expected to emit a
        # 'dtool-config-changed' signal.
        # Assert that the 'dtool-config-changed' signal has been emitted
        mock_emit.assert_called_once_with('dtool-config-changed')


@pytest.mark.asyncio
async def test_do_import_config(running_app, settings):
    """Test do_import_config action."""
    mock_config = {'key1': 'value1', 'key2': 'value2'}
    # Define the mock config file path
    config_file_path = '/path/to/mock/config.json'

    with patch('builtins.open', mock_open(read_data=json.dumps(mock_config))) as mock_file, \
            patch('json.load', return_value=mock_config), \
            patch('dtoolcore.utils.write_config_value_to_file') as mock_write_config, \
            patch.object(running_app, 'emit') as mock_emit:
        # Create a mock for the GLib.Variant object
        mock_value = mock.Mock()
        mock_value.get_string.return_value = config_file_path

        # Perform the import config action
        running_app.do_import_config('import-config', mock_value)

        # Assert that the config file is opened correctly with the specified path
        mock_file.assert_any_call(config_file_path, 'r')

        # Assert that the config values are written to file
        for key, value in mock_config.items():
            mock_write_config.assert_any_call(key, value)

        # Assert that the 'dtool-config-changed' signal has been emitted
        mock_emit.assert_called_once_with('dtool-config-changed')


@pytest.mark.asyncio
async def test_do_export_config(running_app, settings):
    """Test do_export_config action."""
    mock_config = {'key1': 'value1', 'key2': 'value2'}
    # Define the mock config file path
    config_file_path = '/path/to/mock/config.json'

    # Setup mock open
    mock_file_handle = mock_open()
    with patch('builtins.open', mock_file_handle) as mock_file, \
            patch('dtoolcore.utils._get_config_dict_from_file', return_value=mock_config), \
            patch.object(running_app, 'emit') as mock_emit:
        # Create a mock for the GLib.Variant object
        mock_value = mock.Mock()
        mock_value.get_string.return_value = config_file_path

        # Perform the export config action
        running_app.do_export_config('export-config', mock_value)

        # Assert that the config file is opened correctly with the specified path
        mock_file.assert_called_once_with(config_file_path, 'w')

        # Concatenate all the calls to write and compare with expected JSON
        written_data = ''.join(call_arg.args[0] for call_arg in mock_file_handle().write.call_args_list)
        expected_file_content = json.dumps(mock_config, indent=4)
        assert written_data == expected_file_content


@pytest.mark.asyncio
async def test_do_renew_token(running_app):
    """Test do_renew_token action."""
    import asyncio

    with patch('asyncio.create_task') as mock_create_task, \
         patch.object(running_app, 'emit') as mock_emit:  # Patch the emit method

        # Mock values for username, password, and auth_url
        username = "testuser"
        password = "testpass"
        auth_url = "http://auth.example.com"

        # Tuple variant to mimic the value parameter in the method
        value = MagicMock()
        value.unpack.return_value = (username, password, auth_url)

        # Perform the renew token action
        running_app.do_renew_token(None, value)

        # Assert that asyncio.create_task was called with the retrieve_token coroutine
        mock_create_task.assert_called_once()
        args, kwargs = mock_create_task.call_args
        retrieve_token_coro = args[0]
        assert isinstance(retrieve_token_coro, types.CoroutineType)

        # Assert that the coroutine was called with the correct arguments
        assert retrieve_token_coro.cr_code.co_name == 'retrieve_token'
        assert retrieve_token_coro.cr_frame.f_locals['auth_url'] == auth_url
        assert retrieve_token_coro.cr_frame.f_locals['username'] == username
        assert retrieve_token_coro.cr_frame.f_locals['password'] == password

        # The 'do_renew_token' action is expected to emit a
        # 'dtool-config-changed' signal after successful token renewal.
        # Since the actual token retrieval is asynchronous and mocked,
        # we cannot assert this directly in this test.
        # mock_emit.assert_called_once_with('dtool-config-changed')


#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from gi.repository import Gtk


class DtoolSearchResultsRow(Gtk.ListBoxRow):
    __gtype_name__ = 'DtoolSearchResultsRow'

    _margin = 3

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._task = None
        self._search_results = None

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, margin_top=self._margin, margin_bottom=self._margin,
                       margin_start=self._margin, margin_end=self._margin)
        image = Gtk.Image.new_from_icon_name('edit-find-symbolic', Gtk.IconSize.BUTTON)
        image.set_padding(12, 12)
        hbox.pack_start(image, False, False, 0)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        label = Gtk.Label(xalign=0)
        label.set_markup(f'<b>Lookup server</b>')
        vbox.pack_start(label, True, True, 0)
        self._info_label = Gtk.Label(xalign=0)
        self._info_label.set_text('---')
        vbox.pack_start(self._info_label, True, True, 0)
        hbox.pack_start(vbox, True, True, 0)
        self._spinner = Gtk.Spinner(margin_top=self._margin, margin_bottom=self._margin,
                                    margin_start=self._margin, margin_end=self._margin)
        hbox.pack_start(self._spinner, False, False, 0)
        self.add(hbox)

    @property
    def search_results(self):
        return self._search_results

    @search_results.setter
    def search_results(self, search_results):
        self._search_results = search_results

    @property
    def info_label(self):
        return self._info_label

    def start_spinner(self):
        self._spinner.start()

    def stop_spinner(self):
        self._spinner.stop()

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, task):
        self._task = task



import os
from subprocess import Popen

from setuptools import build_meta as _orig
from setuptools.build_meta import *

SOURCE_PATH = os.path.join(os.path.dirname(__file__), os.path.pardir, "dtool_lookup_gui")
TARGET_PATH = os.path.join(os.path.dirname(__file__), os.path.pardir, "dtool_lookup_gui")


def glib_compile_schemas(source_path, target_path):
    if not os.path.exists(target_path):
        os.makedirs(target_path)

    print(f"running glib-compile-schemas {source_path} --targetdir {target_path}")
    Popen(["glib-compile-schemas", source_path, "--targetdir", target_path],
          ).wait()


def build_wheel(wheel_directory, config_settings=None, metadata_directory=None):
    glib_compile_schemas(SOURCE_PATH, TARGET_PATH)
    return _orig.build_wheel(wheel_directory,
                             config_settings=config_settings,
                             metadata_directory=metadata_directory)


def build_sdist(sdist_directory, config_settings=None):
    glib_compile_schemas(SOURCE_PATH, TARGET_PATH)
    return _orig.build_sdist(sdist_directory, config_settings=config_settings)


#
# Copyright 2022 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import platform
import subprocess

from dtoolcore.utils import generous_parse_uri

from ..utils.environ import TemporaryOSEnviron


from gi.repository import Gio


logger = logging.getLogger(__name__)


def launch_default_app_for_uri(uri):
    """Launch OS- and environment-dependent default application for URI."""
    parsed_uri = generous_parse_uri(uri)
    if parsed_uri.scheme != 'file':
        logger.warning("Only meant to open local resources. URI scheme '%s' not supported.", parsed_uri.scheme)

    filepath = parsed_uri.path
    logger.debug("URI '%s' corresponds to local path '%s'.", uri, filepath)

    # This workaround for launching external system applications from
    # within a frozen app as suggested on
    # https://pyinstaller.readthedocs.io/en/stable/runtime-information.html#ld-library-path-libpath-considerations
    # likely not be necessary here. It may, however, be necessary for
    # opening arbitrary file types.
    lp_key = 'LD_LIBRARY_PATH'
    lp_mod = os.environ.get(lp_key + '_ORIG', None)
    if lp_mod is not None:
        logger.debug("Modfify environment with '%s=%s.'", lp_key, lp_mod)
    else:
        lp_mod = os.environ.get(lp_key, "")
    with TemporaryOSEnviron(env={lp_key: lp_mod}):
        logger.debug("Open %s.", uri)
        if platform.system() == 'Darwin':  # macOS
            logger.debug("On macOS, launch 'open %s'", filepath)
            subprocess.call(('open', filepath))
        elif platform.system() == 'Windows':
            if filepath[0] == '/':
                # remove leading slash from path such as /C:/Users/admin/...
                filepath = filepath[1:]
            # convert / to \
            filepath = os.path.abspath(filepath)
            logger.debug("On Windows, launch wth os.startfile('%s')", filepath)
            os.startfile(filepath)
            # direct call of start won't work for directories
            # subprocess.call(('start', filepath))
        else:  # linux variants
            logger.debug("On Linux, launch with Gio.AppInfo.launch_default_for_uri_async('%s')", uri)
            # On Linux, the Gio.AppInfo mechanism fails for opening files, need to investigate
            # Gio.AppInfo.launch_default_for_uri_async(uri)
            # use xdg-open for now, 2022-05-25
            subprocess.call(('xdg-open', filepath))


#
# Copyright 2022 Johannes Laurin Hörmann
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging
import os

import dtoolcore

from gi.repository import Gio, GLib, Gtk

from ..utils.about import pretty_version_text
from ..utils.logging import _log_nested


logger = logging.getLogger(__name__)


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/about_dialog.ui')
class AboutDialog(Gtk.Window):
    __gtype_name__ = 'DtoolAboutDialog'

    version_info_label = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        version_info = pretty_version_text()
        _log_nested(logger.info, version_info)
        self.version_info_label.set_markup(version_info)

    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        """Don't delete, just hide."""
        return self.hide_on_delete()



#
# Copyright 2023 Ashwin Vazhappilly
#           2023 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import asyncio
import logging
import os
import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from dtool_lookup_api.asynchronous import get_versions
from ..utils.logging import _log_nested

# Set up logger for this module
logger = logging.getLogger(__name__)


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/server_versions_dialog.ui')
class ServerVersionsDialog(Gtk.Window):
    __gtype_name__ = 'ServerVersionsDialog'
    server_versions_label = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Apply margins to the server versions label for better aesthetics
        self.server_versions_label.set_margin_start(10)
        self.server_versions_label.set_margin_end(10)
        self.server_versions_label.set_margin_top(10)
        self.server_versions_label.set_margin_bottom(10)

    async def _retrieve_versions(self):
        """Asynchronously fetch server versions and update the label with formatted information."""
        server_versions = await get_versions()
        version_info = self._format_server_versions(server_versions)
        _log_nested(logger.info, version_info)
        self.server_versions_label.set_markup(version_info)

    def _format_server_versions(self, server_versions):
        """Format server versions, sorting components by name length."""
        sorted_components = sorted(server_versions.keys(), key=len)

        # Use list comprehension for concise formatting of server versions
        formatted_versions = [f"{component}: <b>{server_versions[component]}</b>" for component in sorted_components]
        return "\n".join(formatted_versions)

    @Gtk.Template.Callback()
    def on_show(self, widget):
        """Callback executed when the window is shown; fetches server versions."""
        asyncio.create_task(self._retrieve_versions())

    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        """Callback executed when window close event is triggered; hides the window instead of deleting."""
        return self.hide_on_delete()



#
# Copyright 2021-2022 Johannes Laurin Hörmann
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import logging
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '4')
from gi.repository import GLib, GObject, Gio, Gtk, GtkSource, GdkPixbuf

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

gio_appinfo_list = Gio.AppInfo.get_all()

# print(app_info_list)

for gio_appinfo in gio_appinfo_list:
    gio_appinfo_dict = {
        'commandline': gio_appinfo.get_commandline(),
        'description': gio_appinfo.get_description(),
        'display_name': gio_appinfo.get_display_name(),
        'executable': gio_appinfo.get_executable(),
        'id': gio_appinfo.get_id(),
        'supported_types': gio_appinfo.get_supported_types(),
    }
    logger.info(gio_appinfo_dict)
# Gio.AppInfo.launch_default_for_uri_async(uri)


class ArgumentDefaultsAndRawDescriptionHelpFormatter(
    argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass

parser = argparse.ArgumentParser(description=__doc__,
                                 formatter_class=ArgumentDefaultsAndRawDescriptionHelpFormatter)

parser.add_argument('uri', help='URI to launch')


args = parser.parse_args()

uri = args.uri

scheme = GLib.uri_parse_scheme(uri)
logger.info("URI %s has scheme %s", uri, scheme)


gio_appinfo = Gio.AppInfo.get_default_for_uri_scheme(scheme)
gio_appinfo_dict = {
    'commandline': gio_appinfo.get_commandline(),
    'description': gio_appinfo.get_description(),
    'display_name': gio_appinfo.get_display_name(),
    'executable': gio_appinfo.get_executable(),
    'id': gio_appinfo.get_id(),
    'supported_types': gio_appinfo.get_supported_types(),
}
logger.info(gio_appinfo_dict)

logger.info("Launch default app for %s", uri)
Gio.AppInfo.launch_default_for_uri_async(uri)



#
# Copyright 2020-2022 Johannes Laurin Hörmann
#           2020-2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import json
import logging

from aiohttp.client_exceptions import ContentTypeError

from dtool_lookup_gui import is_uuid
from dtool_lookup_gui.models.simple_graph import SimpleGraph
from dtool_lookup_gui.models.search_state import SearchState

logger = logging.getLogger(__name__)


def _log_nested(log_func, dct):
    for l in json.dumps(dct, indent=2, default=str).splitlines():
        log_func(l)


class DependencyGraph:
    def __init__(self):
        self._search_state = SearchState()
        self._reset_graph()
        self._reset_graph()

    @property
    def graph(self):
        return self._graph

    def _reset_graph(self):
        self._graph = SimpleGraph()
        self._uuid_to_vertex = {}
        self._missing_uuids = []

    async def trace_dependencies(self, lookup, root_uuid, dependency_keys=None):
        """Build dependency graph by UUID."""
        logger.debug(f"Build dependency graph for root '{root_uuid}'.")
        self._reset_graph()

        if isinstance(dependency_keys, str):
            dependency_keys = json.loads(dependency_keys)

        if (dependency_keys is not None) and (not isinstance(dependency_keys, list)):
            logger.warning("Dependency keys not valid. Ignored.")
            dependency_keys = None

        pagination = {}
        try:
            datasets = await lookup.get_graph_by_uuid(uuid=root_uuid,
                                                      dependency_keys=dependency_keys,
                                                      page_size=self._search_state.page_size,
                                                      pagination=pagination)
        except ContentTypeError as exc:
            logger.error("%s", exc)
            return

        logger.debug("Got first batch of dependency graph datasets with pagination information '%s'.", pagination)

        self._search_state.ingest_pagination_information(pagination=pagination)

        for page in range(self._search_state.first_page+1, self._search_state.last_page+1):
            try:
                more_datasets = await lookup.get_graph_by_uuid(uuid=root_uuid,
                                                               dependency_keys=dependency_keys,
                                                               page_number=page, page_size=self._search_state.page_size,
                                                               pagination=pagination)
            except ContentTypeError as exc:
                logger.error("%s", exc)
                return
            logger.debug("Got batch #%s of dependency graph datasets with pagination information '%s'.", page, pagination)
            if len(more_datasets) > 0:
                datasets.extend(more_datasets)

        logger.debug("Server response on querying dependency graph for UUID = {}.".format(root_uuid))
        _log_nested(logger.debug, datasets)

        for dataset in datasets:
            # This check should be redundant, as all documents have field 'uuid'
            # and this field is unique:
            uuid = dataset['uuid']
            if 'uuid' in dataset and uuid not in self._uuid_to_vertex:
                logger.debug(f"Add dependency graph vertex '{uuid}'.")
                v = self._graph.add_vertex(
                    uuid=uuid,
                    name=dataset['name'],
                    kind='root' if uuid == root_uuid else 'dependent')
                self._uuid_to_vertex[uuid] = v

        for dataset in datasets:
            if 'uuid' in dataset and 'derived_from' in dataset:
                for parent_uuid in dataset['derived_from']:
                    if is_uuid(parent_uuid):
                        if parent_uuid not in self._uuid_to_vertex:
                            # This UUID is present in the graph but not in the database
                            logger.debug(f"Add dependency graph vertex of missing dataset '{parent_uuid}'.")
                            v = self._graph.add_vertex(
                                uuid=parent_uuid,
                                name='Dataset does not exist in database.',
                                kind='does-not-exist')
                            self._uuid_to_vertex[parent_uuid] = v
                            self._missing_uuids += [parent_uuid]

                        logger.debug(f"Add dependency graph edge between child '{dataset['uuid']}' and parent {parent_uuid}.")
                        self._graph.add_edge(
                            self._uuid_to_vertex[dataset['uuid']],
                            self._uuid_to_vertex[parent_uuid])
                    else:
                        logger.warning(
                            "Parent dataset '{}' of child '{}': '{}' is no "
                            "valid UUID, ignored.".format(parent_uuid,
                                                          dataset['uuid'],
                                                          dataset['name']))
        logger.debug(f"Done building dependency graph for root '{root_uuid}'.")

    @property
    def missing_uuids(self):
        return self._missing_uuids


#
# Copyright 2023 Ashwin Vazhappilly
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import asyncio
import logging
import os
import gi
import json

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from dtool_lookup_api.asynchronous import get_config

# Set up logger for this module
logger = logging.getLogger(__name__)

@Gtk.Template(filename=f'{os.path.dirname(__file__)}/config_details.ui')
class ConfigDialog(Gtk.Window):
    __gtype_name__ = 'ConfigDialog'
    config_text_view = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Setting margins for better spacing
        self.config_text_view.set_left_margin(30)
        self.config_text_view.set_right_margin(10)
        self.config_text_view.set_top_margin(5)
        self.config_text_view.set_bottom_margin(5)

    async def _retrieve_config(self):
        """Asynchronously fetch server configuration and update the text view."""
        server_config = await get_config()
        config_info = self._format_server_config(server_config)
        buffer = self.config_text_view.get_buffer()
        buffer.set_text("")  # Clearing the buffer
        buffer.insert_at_cursor("\n".join(config_info))

    def _format_server_config(self, server_config):
        """Format the server configuration into a human-readable string representation using json.dumps."""
        formatted_config = json.dumps(server_config, indent=4)
        return formatted_config.splitlines()

    @Gtk.Template.Callback()
    def on_config_show(self, widget):
        """Callback executed when the window is shown; fetches server config."""
        asyncio.create_task(self._retrieve_config())

    @Gtk.Template.Callback()
    def on_config_delete(self, widget, event):
        """Callback executed when window close event is triggered; hides the window instead of deleting."""
        return self.hide_on_delete()



#
# Copyright 2021-2023 Johannes Laurin Hörmann
#           2023 Ashwin Vazhappilly
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import asyncio
import datetime
import logging
import os

import dtoolcore

from gi.repository import Gio, GLib, Gtk

from dtool_lookup_api.core.config import Config

from ..models.settings import settings
from .authentication_dialog import AuthenticationDialog
from .s3_configuration_dialog import S3ConfigurationDialog
from .smb_configuration_dialog import SMBConfigurationDialog


_DTOOL_CONFIG_PREFIXES = {
    'DTOOL_S3_ENDPOINT_': 's3',
    'DTOOL_SMB_SERVER_NAME_': 'smb',
}

_DTOOL_README_TEMPLATE_FPATH_KEY = "DTOOL_README_TEMPLATE_FPATH"
_DTOOL_USER_FULL_NAME_KEY = "DTOOL_USER_FULL_NAME"
_DTOOL_USER_EMAIL_KEY = "DTOOL_USER_EMAIL"


logger = logging.getLogger(__name__)


def _get_base_uri(key):
    for prefix, schema in _DTOOL_CONFIG_PREFIXES.items():
        if key.startswith(prefix):
            return f'{schema}://{key[len(prefix):]}'
    return None


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/settings_dialog.ui')
class SettingsDialog(Gtk.Window):
    __gtype_name__ = 'DtoolSettingsDialog'

    lookup_url_entry = Gtk.Template.Child()
    token_entry = Gtk.Template.Child()
    authenticator_url_entry = Gtk.Template.Child()
    renew_token_button = Gtk.Template.Child()
    dependency_keys_entry = Gtk.Template.Child()
    verify_ssl_certificate_switch = Gtk.Template.Child()
    disable_authentication_switch = Gtk.Template.Child()
    base_uris_list_box = Gtk.Template.Child()

    dtool_user_full_name_entry = Gtk.Template.Child()
    dtool_user_email_entry = Gtk.Template.Child()
    dtool_readme_template_fpath_file_chooser_button = Gtk.Template.Child()

    item_download_directory_file_chooser_button = Gtk.Template.Child()
    choose_item_download_target_directory_checkbox = Gtk.Template.Child()
    open_downloaded_item_checkbox = Gtk.Template.Child()
    yaml_linting_switch = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        settings.settings.bind("dependency-keys", self.dependency_keys_entry, 'text', Gio.SettingsBindFlags.DEFAULT)

        settings.settings.bind("choose-item-download-directory",
                               self.choose_item_download_target_directory_checkbox,
                               'active', Gio.SettingsBindFlags.DEFAULT)
        settings.settings.bind("open-downloaded-item",
                               self.open_downloaded_item_checkbox,
                               'active', Gio.SettingsBindFlags.DEFAULT)
        settings.settings.bind("yaml-linting-enabled",
                               self.yaml_linting_switch,
                               'active', Gio.SettingsBindFlags.DEFAULT)


        # register own refresh method as listener for app-central dtool-config-changed signal
        self.get_application().connect("dtool-config-changed", self.on_dtool_config_changed)

        self._refresh_settings_dialog()

    def _refresh_settings_dialog(self):
        logger.debug("Refresh settings dialog.")

        # access lookup config via lookup api
        if Config.lookup_url is not None:
            logger.debug(f"Current lookup server url: {Config.lookup_url}")
            self.lookup_url_entry.set_text(Config.lookup_url)
        else:
            logger.debug("No lookup server url configured.")
            self.lookup_url_entry.set_text('')

        if Config.token is not None:
            logger.debug(f"Current lookup server token: {Config.token}")
            self.token_entry.set_text(Config.token)
        else:
            logger.debug("No lookup server token configured.")
            self.token_entry.set_text('')

        if Config.auth_url is not None:
            logger.debug(f"Current lookup server auth url: {Config.auth_url}")
            self.authenticator_url_entry.set_text(Config.auth_url)
        else:
            logger.debug("No lookup server auth url configured.")
            self.authenticator_url_entry.set_text('')

        if Config.verify_ssl is not None:
            logger.debug(f"Current lookup server verify ssl: {Config.verify_ssl}")
            # following line throws segmentation fault
            self.verify_ssl_certificate_switch.set_state(Config.verify_ssl)
        else:
            logger.debug("No lookup server verify ssl configured, set True.")
            self.verify_ssl_certificate_switch.set_state(True)

        if Config.disable_authentication is not None:
            logger.debug(f"Current lookup server disable authentication: {Config.disable_authentication}")
            self.disable_authentication_switch.set_state(Config.disable_authentication)
        else:
            logger.debug("No lookup server disable authentication configured, set False.")
            self.disable_authentication_switch.set_state(False)

        disabled = self.disable_authentication_switch.get_state()
        self.authenticator_url_entry.set_sensitive(not disabled)
        self.renew_token_button.set_sensitive(not disabled)
        self.token_entry.set_sensitive(not disabled)

        # access basic config via default dtool config
        self.dtool_user_full_name_entry.set_text(
            dtoolcore.utils.get_config_value(_DTOOL_USER_FULL_NAME_KEY, default=""))

        self.dtool_user_email_entry.set_text(
            dtoolcore.utils.get_config_value(_DTOOL_USER_EMAIL_KEY, default=""))

        dtool_readme_template_fpath = dtoolcore.utils.get_config_value(_DTOOL_README_TEMPLATE_FPATH_KEY)
        if dtool_readme_template_fpath is not None:
            logger.debug(f"Current readme template: {dtool_readme_template_fpath}")
            self.dtool_readme_template_fpath_file_chooser_button.set_filename(dtool_readme_template_fpath)

        # get gio settings (if not bound)
        if os.path.isdir(settings.item_download_directory):
            logger.debug(f"Current default item download directory: %s",  settings.item_download_directory)
            self.item_download_directory_file_chooser_button.set_filename(settings.item_download_directory)

        logger.debug("Refresh list of endpoints.")
        asyncio.create_task(self._refresh_list_of_endpoints())

    async def _refresh_list_of_endpoints(self):
        logger.debug("Refresh base uris list box.")
        await self.base_uris_list_box.refresh(on_configure=self.on_configure_base_uri_clicked, local=False,
                                              search_results=False)

        logger.debug("Add button for new end points.")
        # Plus button for adding new endpoints
        row = Gtk.ListBoxRow()
        image = Gtk.Image.new_from_icon_name('list-add-symbolic', Gtk.IconSize.BUTTON)
        image.set_margin_top(20)
        image.set_margin_bottom(20)
        image.set_margin_start(12)
        image.set_margin_end(12)
        row.connect('state-changed', self.on_base_uri_state_changed)
        row.add(image)
        self.base_uris_list_box.add(row)

        logger.debug("base uris list box show all.")
        self.base_uris_list_box.show_all()
        logger.debug("Done refreshing settings dialog.")

    def on_dtool_config_changed(self, widget):
        """Signal handler for dtool-config-changed."""
        self._refresh_settings_dialog()

    # signal handlers
    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        # Write back lookup configuration via lookup api
        Config.lookup_url = self.lookup_url_entry.get_text()
        Config.token = self.token_entry.get_text()
        Config.auth_url = self.authenticator_url_entry.get_text()
        Config.verify_ssl = self.verify_ssl_certificate_switch.get_state()
        Config.disable_authentication = self.disable_authentication_switch.get_state()

        # write back basic config via default dtool api
        dtool_user_full_name = self.dtool_user_full_name_entry.get_text()
        if dtool_user_full_name != dtoolcore.utils.get_config_value(_DTOOL_USER_FULL_NAME_KEY, default=""):
            logger.debug(f"{_DTOOL_USER_FULL_NAME_KEY} changed to {dtool_user_full_name}, write to config.")
            dtoolcore.utils.write_config_value_to_file(_DTOOL_USER_FULL_NAME_KEY, dtool_user_full_name)

        dtool_user_email = self.dtool_user_email_entry.get_text()
        if dtool_user_email != dtoolcore.utils.get_config_value(_DTOOL_USER_EMAIL_KEY, default=""):
            logger.debug(f"{_DTOOL_USER_EMAIL_KEY} changed to {dtool_user_email}, write to config.")
            dtoolcore.utils.write_config_value_to_file(_DTOOL_USER_EMAIL_KEY, dtool_user_email)

        dtool_readme_template_fpath = self.dtool_readme_template_fpath_file_chooser_button.get_filename()
        if (dtool_readme_template_fpath is not None
                and dtool_readme_template_fpath != dtoolcore.utils.get_config_value(_DTOOL_README_TEMPLATE_FPATH_KEY)):
            logger.debug(f"{_DTOOL_README_TEMPLATE_FPATH_KEY} changed to {dtool_readme_template_fpath}, write to config.")
            dtoolcore.utils.write_config_value_to_file(_DTOOL_README_TEMPLATE_FPATH_KEY, dtool_readme_template_fpath)

        return self.hide_on_delete()

    @Gtk.Template.Callback()
    def on_renew_token_clicked(self, widget):
        # show authentication dialogue and get username and password
        def authenticate(username, password):
            auth_url = self.authenticator_url_entry.get_text()
            user_pass_auth_variant = GLib.Variant.new_tuple(GLib.Variant.new_string(username),
                                                            GLib.Variant.new_string(password),
                                                            GLib.Variant.new_string(auth_url))
            self.get_action_group("app").activate_action('renew-token', user_pass_auth_variant)

        AuthenticationDialog(authenticate, Config.username, Config.password).show()

    @Gtk.Template.Callback()
    def on_reset_config_clicked(self, widget):
        """Process clicked signal from reset-config button."""
        self.get_action_group("app").activate_action('reset-config')

    @Gtk.Template.Callback()
    def on_import_config_clicked(self, widget):
        """Process clicked signal from import-config button."""
        dialog = Gtk.FileChooserDialog(title=f"Import dtool config from file", parent=self,
                                       action=Gtk.FileChooserAction.OPEN)
        dialog.add_buttons(Gtk.STOCK_CANCEL,
                           Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OPEN,
                           Gtk.ResponseType.OK)
        dialog.set_select_multiple(False)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            dest_filename = dialog.get_filename()
            self.get_action_group("app").activate_action('import-config', GLib.Variant.new_string(dest_filename))

        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()

    @Gtk.Template.Callback()
    def on_export_config_clicked(self, widget):
        """Process clicked signal from import-config button."""
        dialog = Gtk.FileChooserDialog(title=f"Export dtool config to file", parent=self,
                                       action=Gtk.FileChooserAction.SAVE)
        dialog.add_buttons(Gtk.STOCK_CANCEL,
                           Gtk.ResponseType.CANCEL,
                           Gtk.STOCK_OK,
                           Gtk.ResponseType.OK)
        suggested_file_name = f"{datetime.datetime.now().isoformat()}-{self.get_application().get_application_id()}-dtool.json"
        dialog.set_current_name(suggested_file_name)
        dialog.set_do_overwrite_confirmation(True)

        response = dialog.run()

        if response == Gtk.ResponseType.OK:
            dest_filename = dialog.get_filename()
            self.get_action_group("app").activate_action('export-config', GLib.Variant.new_string(dest_filename))

        elif response == Gtk.ResponseType.CANCEL:
            pass
        dialog.destroy()

    # binding the filechooser to settings via property isn't possible, hence work with signals here
    @Gtk.Template.Callback()
    def on_item_download_directory_file_chooser_button_file_set(self, widget):
        item_download_directory = widget.get_file().get_path()
        if item_download_directory is None:
            logger.error("Selected directory invalid.")
            return

        logger.debug("Selected default item download directory '%s'.", item_download_directory)
        settings.item_download_directory = item_download_directory

    _configuration_dialogs = {
        's3': S3ConfigurationDialog,
        'smb': SMBConfigurationDialog,
    }

    def on_configure_base_uri_clicked(self, widget):
        base_uri = widget.get_parent().get_parent().base_uri
        self._configuration_dialogs[base_uri.scheme](
            lambda: asyncio.create_task(self._refresh_list_of_endpoints()), base_uri.uri_name).show()

    def on_base_uri_state_changed(self, widget, state):
        if state == Gtk.StateType.ACTIVE:
            S3ConfigurationDialog(lambda: asyncio.create_task(self._refresh_list_of_endpoints())).show()

    @Gtk.Template.Callback()
    def on_yaml_linting_switch_state_set(self, widget, state):
        """Switch yaml linting on and off"""
        settings.yaml_linting_enabled = state

    @Gtk.Template.Callback()
    def on_verify_ssl_certificate_switch_state_set(self, widget, state):
        Config.verify_ssl = state

    @Gtk.Template.Callback()
    def on_disable_authentication_switch_state_set(self, widget, state):
        Config.disable_authentication = state
        self.authenticator_url_entry.set_sensitive(not state)
        self.renew_token_button.set_sensitive(not state)
        self.token_entry.set_sensitive(not state)


#
# Copyright 2022 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os

from .logging import _log_nested


class TemporaryOSEnviron:
    """Preserve original os.environ context manager."""

    def __init__(self, env=None):
        """env is a flat dict to be inserted into os.environ."""
        self._insertions = env

    def __enter__(self):
        """Store backup of current os.environ."""
        logger = logging.getLogger(__name__)
        logger.debug("Backed-up os.environ:")
        _log_nested(logger.debug, dict(os.environ))
        self._original_environ = dict(os.environ.copy())

        if self._insertions:
            for k, v in self._insertions.items():
                logger.debug("Inject env var '{}' = '{}'".format(k, v))
                os.environ[k] = str(v)

        logger.debug("Initial modified os.environ:")
        _log_nested(logger.debug, dict(os.environ))

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Restore backed up os.environ."""
        logger = logging.getLogger(__name__)
        os.environ.clear()
        os.environ.update(self._original_environ)
        logger.debug("Recovered os.environ:")
        _log_nested(logger.debug, dict(os.environ))



class SearchState:
    """The client model of the search state, e.g. current pagination, sorting, search keywords, ..."""

    def __init__(self):

        self._search_text = ""

        self._page_size = 10
        self._fetching_results = False

        self.reset_pagination()
        self.reset_sorting()

        self.sort_field_label_dict = {
            "uri": "URI",
            "name": "name",
            "base_uri": "base URI",
            "created_at": "created at",
            "frozen_at": "frozen at",
            "uuid": "UUID",
            "creator_username": "creator username",
        }
        self.label_sort_field_dict = {value: key for key, value in self.sort_field_label_dict.items()}
        self.page_size_choices = [5, 10, 20, 50, 100]

    def reset_pagination(self):
        """Reset pagination information to single page"""
        self._current_page = 1
        self._first_page = 1
        self._last_page = 1
        self._total_pages = 1
        self._total_number_of_entries = 0

    def reset_sorting(self):
        """Reset sorting information to uri"""
        self._sort_fields = ["uri"]
        self._sort_order = [1]

    # search_text property
    def get_search_text(self):
        return self._search_text

    def set_search_text(self, value):
        self._search_text = value

    # current_page property
    def get_current_page(self):
        return self._current_page

    def set_current_page(self, value):
        # if not isinstance(value, int) or value < 1:
        #    raise ValueError("Current page must be a positive integer")
        if value < 1:
            value = 1
        elif value > self.last_page:
            value = self.last_page
        self._current_page = value

    # last_page property
    def get_last_page(self):
        return self._last_page

    def set_last_page(self, value):
        if not isinstance(value, int) or value < 1:
            raise ValueError("Last page must be a positive integer")
        self._last_page = value

    # first_page property
    def get_first_page(self):
        return self._first_page

    def set_first_page(self, value):
        if not isinstance(value, int) or value < 1:
            raise ValueError("Last page must be a positive integer")
        self._first_page = value

    # page_size property
    def get_page_size(self):
        return self._page_size

    def set_page_size(self, value):
        if not isinstance(value, int) or value < 1:
            raise ValueError("Page size must be a positive integer")
        self._page_size = value

    # fetching_results property
    def get_fetching_results(self):
        return self._fetching_results

    def set_fetching_results(self, value):
        if not isinstance(value, bool):
            raise ValueError("Fetching results must be a boolean")
        self._fetching_results = value

    # sort_fields property
    def get_sort_fields(self):
        return self._sort_fields

    def set_sort_fields(self, value):
        if isinstance(value, str):
            value = [value]
        if not isinstance(value, list):
            raise ValueError("Sort fields must be a list")
        for sort_field in value:
            if sort_field not in self.sort_field_label_dict:
                raise ValueError("Sort field %s not allowed", sort_field)
        self._sort_fields = value

    # sort_order property
    def get_sort_order(self):
        return self._sort_order

    def set_sort_order(self, value):
        if isinstance(value, int):
            value = [value]
        if not isinstance(value, list):
            raise ValueError("Sort order must be a list")
        if len(value) != len(self.sort_fields):
            raise ValueError("Sort order list must have same length as sort fields list")
        self._sort_order = value

    def ingest_pagination_information(self, pagination):
        """Sets pagination information to values found in dict of format

           {"total": 284, "total_pages": 29, "first_page": 1, "last_page": 29, "page": 1, "next_page": 2}
        """
        self.first_page = pagination.get("first_page", 1)
        self.last_page = pagination.get("last_page", 1)
        self.current_page = pagination.get("page", 1)
        self._total_pages = pagination.get("total_pages", 1)
        self._total_number_of_entries = pagination.get("total", 0)

    def ingest_sorting_information(self, sorting):
        """Sets sorting information to values found in dict of format

               {"base_uri": 1, "name": -1}
        """
        sort_fields = []
        sort_order = []

        sort_info = sorting.get("sort", {})
        for key, value in sort_info.items():
            sort_fields.append(key)
            sort_order.append(value)

        self.sort_fields = sort_fields
        self.sort_order = sort_order

    # Properties
    search_text = property(get_search_text, set_search_text)
    current_page = property(get_current_page, set_current_page)
    last_page = property(get_last_page, set_last_page)
    first_page = property(get_first_page, set_first_page)
    page_size = property(get_page_size, set_page_size)
    fetching_results = property(get_fetching_results, set_fetching_results)
    sort_fields = property(get_sort_fields, set_sort_fields)
    sort_order = property(get_sort_order, set_sort_order)

    @property
    def next_page(self):
        if self.current_page >= self.last_page:
            return self.last_page
        else:
            return self.current_page + 1

    @property
    def previous_page(self):
        if self.current_page <= self.first_page:
            return self.first_page
        else:
            return self.current_page - 1
    @property
    def total_pages(self):
        return self._total_pages

    @property
    def total_number_of_entries(self):
        return self._total_number_of_entries


#
# Copyright 2023 Johannes Laurin Hörmann
#           2023 Ashwin Vazhappilly
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import dtoolcore
from gi.repository import Gio, GLib, Gtk, Gdk
from dtool_lookup_api.core.config import Config
from ..utils.about import pretty_version_text
from ..utils.logging import _log_nested
from .settings_dialog import SettingsDialog

# Initialize logging
logger = logging.getLogger(__name__)

@Gtk.Template(filename=f'{os.path.dirname(__file__)}/login_window.ui')
class LoginWindow(Gtk.Window):
    __gtype_name__ = 'LoginWindow'

    # Map GTK Template Child widgets to Python attributes
    username_entry = Gtk.Template.Child()
    password_entry = Gtk.Template.Child()
    login_button = Gtk.Template.Child()
    skip_button = Gtk.Template.Child()
    settings_button = Gtk.Template.Child()

    # Initialize the LoginWindow instance
    def __init__(self, follow_up_action=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.application = self.get_application()
        self.settings_dialog = SettingsDialog(application=self.application)
        self._follow_up_action = follow_up_action

        # Set the default values from the Config, ensuring they are not None
        if Config.username is not None:
            self.username_entry.set_text(Config.username)
        if Config.password is not None:  # Consider security implications
            self.password_entry.set_text(Config.password)

        self.username_entry.connect("activate", self.on_username_activate)
        self.password_entry.connect("activate", self.on_password_activate)

    def on_username_activate(self, entry):
        self.password_entry.grab_focus()

    def on_password_activate(self, entry):
        self.on_login_button_clicked(None)

    # Handle the 'Login' button click event
    @Gtk.Template.Callback()
    def on_login_button_clicked(self, widget):
        # Fetch entered username and password
        username = self.username_entry.get_text()
        password = self.password_entry.get_text()

        # Create a GLib Variant tuple for authentication
        user_pass_auth_variant = GLib.Variant.new_tuple(
            GLib.Variant.new_string(username),
            GLib.Variant.new_string(password),
            GLib.Variant.new_string(Config.auth_url)
        )

        # Trigger the 'renew-token' action for authentication
        self.get_action_group("app").activate_action('renew-token', user_pass_auth_variant)

        if self._follow_up_action is not None:
            self._follow_up_action()

        self.close()

    # Handle the 'Skip' button click event
    @Gtk.Template.Callback()
    def on_skip_button_clicked(self, widget):
        # Close the login window
        self.close()

    @Gtk.Template.Callback()
    def on_settings_button_clicked_cb(self, widget):
        logger.info("Settings button clicked. Opening settings dialog.")
        self.settings_dialog.show()

    # Handle the window close event
    @Gtk.Template.Callback()
    def on_delete(self, widget, event):
        # Hide the window instead of deleting it
        return self.hide_on_delete()





#
# Copyright 2022 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from pkg_resources import iter_entry_points

import dtoolcore
import dtool_lookup_api

try:
    from dtool import __version__ as dtool_version
except ImportError:
    dtool_version = None

from .. import __version__

# from https://github.com/jic-dtool/dtool-cli/blob/master/dtool_cli/cli.py
def pretty_version_text():
    """Return pretty version text listing all plugins."""
    if dtool_version is not None:
        version_lines = ["dtool, version {}".format(dtool_version)]
    else:
        version_lines = []
    version_lines.append("\n<b>Base</b>:")
    version_lines.append("dtoolcore, version {}".format(dtoolcore.__version__))
    version_lines.append("dtool-lookup-api, version {}".format(dtool_lookup_api.__version__))
    version_lines.append("dtool-lookup-gui, version {}".format(__version__))

    # List the storage broker packages.
    version_lines.append("\n<b>Storage brokers</b>:")
    for ep in iter_entry_points("dtool.storage_brokers"):
        package = ep.module_name.split(".")[0]
        dyn_load_p = __import__(package)
        version = dyn_load_p.__version__
        storage_broker = ep.load()
        version_lines.append(
            "{}, {}, version {}".format(
                storage_broker.key,
                package.replace("_", "-"),
                version))

    # List the plugin packages.
    modules = [ep.module_name for ep in iter_entry_points("dtool.cli")]
    packages = set([m.split(".")[0] for m in modules])
    version_lines.append("\n<b>Plugins</b>:")
    for p in packages:
        dyn_load_p = __import__(p)
        version_lines.append(
            "{}, version {}".format(
                p.replace("_", "-"),
                dyn_load_p.__version__))

    return "\n".join(version_lines)



#
# Copyright 2023 Ashwin Vazhappilly
#           2022-2023 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import asyncio
import time
from pathlib import Path

import dtoolcore
import pytest
from unittest.mock import patch, MagicMock, AsyncMock, Mock
from gi.repository import Gtk, Gio, GLib

from dtool_lookup_gui.views.main_window import MainWindow
@pytest.mark.asyncio
async def test_do_search_direct_call(populated_app_with_mock_data):
    """
    Test the do_search method for directly processing a search in a populated application.
    This test checks if the search action correctly triggers the search process and
    verifies that the search results are appropriately loaded and displayed in the UI.
    """

    # Helper function to wait for datasets to load in the list box.
    async def wait_for_datasets_to_load(list_box, timeout=100):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)  # Yield control to allow other async tasks to run
        return False  # Timeout reached if datasets are not loaded within the specified time

    # Get main window of application
    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action to load datasets
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded in the dataset list box
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Create a GLib.Variant with a test search query
    search_query = "test_search_query"
    search_text_variant = GLib.Variant.new_string(search_query)

    # Create and add the search action to the main window of the application
    search_action = Gio.SimpleAction.new("search", search_text_variant.get_type())
    main_window.add_action(search_action)

    # Connect the search action to the do_search method of the main window
    search_action.connect("activate", main_window.do_search)

    # Trigger the search action with the test search query
    search_action.activate(search_text_variant)

    # Optionally, wait for the search results to be processed and displayed
    # This step depends on the application's implementation and response time
    await asyncio.sleep(1)  # Adjust this sleep duration as needed

    # Perform assertions to verify that the search results are correctly displayed
    assert len(main_window.base_uri_list_box.get_children()) > 0, "No search results found"
    # Additional assertions can be added here based on the expected outcomes of the search

    # Await completion of any remaining asynchronous tasks related to the search
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)


@pytest.mark.asyncio
async def test_do_search_action_trigger(running_app):
    """Test if 'search' action triggers do_search method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Setup necessary mocks for the action trigger
    main_window._search = MagicMock()

    # Create and add the action
    search_text_variant = GLib.Variant.new_string("test_search_query")
    search_action = Gio.SimpleAction.new("search", search_text_variant.get_type())
    main_window.add_action(search_action)

    # Patch do_search method after action is added
    with patch.object(main_window, 'do_search', new_callable=MagicMock) as mock_do_search:
        # Connect the action
        search_action.connect("activate", main_window.do_search)

        # Trigger the action
        search_action.activate(search_text_variant)

        # Assert that do_search was called once
        mock_do_search.assert_called_once_with(search_action, search_text_variant)


@pytest.mark.asyncio
async def test_do_select_dataset_row_by_row_index_direct_call(populated_app_with_mock_data):
    """
    Test the do_select_dataset_row_by_row_index method for directly selecting a dataset row by index.
    It verifies if the dataset list is correctly populated and the specified dataset row is selected.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True
            await asyncio.sleep(0.1)
        return False

    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action and wait for datasets to load
    main_window.activate_action('refresh-view')
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Create a GLib.Variant with a valid test row index (e.g., 0 for the first row)
    row_index = 6
    row_index_variant = GLib.Variant.new_uint32(row_index)

    # Create and add the select dataset action to the main window of the application
    select_dataset_action = Gio.SimpleAction.new("select-dataset", row_index_variant.get_type())
    main_window.add_action(select_dataset_action)

    # Connect the select dataset action to the do_select_dataset_row_by_row_index method
    select_dataset_action.connect("activate", main_window.do_select_dataset_row_by_row_index)

    # Trigger the select dataset action with the test row index
    select_dataset_action.activate(row_index_variant)

    # Optionally, wait for the UI to update if necessary
    await asyncio.sleep(0.1)  # Adjust this sleep duration as needed

    # Perform assertions to verify that the correct dataset row is selected
    selected_row = main_window.dataset_list_box.get_selected_row()

    # Assert that the selected row is not None and has the expected index
    assert selected_row is not None, "No dataset row was selected"
    assert selected_row.get_index() == row_index, f"Expected row index {row_index}, got {selected_row.get_index()}"

    # Await completion of any remaining asynchronous tasks related to the selection
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)


@pytest.mark.asyncio
async def test_do_select_dataset_row_by_row_index_action_trigger(running_app):
    """Test if 'select-dataset' action triggers do_select_dataset_row_by_row_index method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Mock dependencies
    mock_dataset_list_box = MagicMock()
    main_window.dataset_list_box = mock_dataset_list_box

    # Setup necessary mocks for the action trigger
    main_window._select_dataset_row_by_row_index = MagicMock()

    # Create and add the action
    row_index_variant = GLib.Variant.new_uint32(1)  # Example index
    select_dataset_action = Gio.SimpleAction.new("select-dataset", row_index_variant.get_type())
    main_window.add_action(select_dataset_action)

    # Patch do_select_dataset_row_by_row_index method after action is added
    with patch.object(main_window, 'do_select_dataset_row_by_row_index',
                      new_callable=MagicMock) as mock_do_select_dataset_row_by_row_index:
        # Connect the action
        select_dataset_action.connect("activate", main_window.do_select_dataset_row_by_row_index)

        # Trigger the action
        select_dataset_action.activate(row_index_variant)

        # Assert that do_select_dataset_row_by_row_index was called once
        mock_do_select_dataset_row_by_row_index.assert_called_once_with(select_dataset_action, row_index_variant)


@pytest.mark.asyncio
async def test_do_show_dataset_details_by_uri_direct_call(populated_app_with_mock_data):
    """
    Test the do_show_dataset_details_by_uri method for processing a URI directly. It verifies
    if the correct dataset is selected and its content is displayed based on the URI.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)  # Yield control to allow other async tasks to run
        return False  # Timeout reached

    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Define a URI that corresponds to a dataset in the populated dataset list
    test_uri = "s3://test-bucket/1a1f9fad-8589-413e-9602-5bbd66bfe675"

    # Create a GLib.Variant with the test URI
    uri_variant = GLib.Variant.new_string(test_uri)

    # Call the do_show_dataset_details_by_uri method with the test URI
    main_window.do_show_dataset_details_by_uri(None, uri_variant)

    # Retrieve the dataset that should be selected based on the URI
    index = main_window.dataset_list_box.get_row_index_from_uri(test_uri)
    selected_row = main_window.dataset_list_box.get_row_at_index(index)
    selected_dataset = selected_row.dataset if selected_row is not None else None

    # Assert that the dataset with the given URI is selected
    assert selected_dataset is not None, "No dataset was selected"
    assert selected_dataset.uri == test_uri, f"Expected URI {test_uri}, got {selected_dataset.uri}"

    # Await completion of all tasks related to the test
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)


@pytest.mark.asyncio
async def test_do_show_dataset_details_by_uri_action_trigger(running_app):
    """Test if 'show-dataset-by-uri' action triggers do_show_dataset_details_by_uri method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Mock dependencies
    mock_dataset_list_box = MagicMock()
    main_window.dataset_list_box = mock_dataset_list_box

    # Setup necessary mocks for the action trigger
    main_window._show_dataset_details_by_uri = MagicMock()

    # Create and add the action
    uri_variant = GLib.Variant.new_string("dummy_uri")
    show_dataset_by_uri_action = Gio.SimpleAction.new("show-dataset-by-uri", uri_variant.get_type())
    main_window.add_action(show_dataset_by_uri_action)

    # Patch do_show_dataset_details_by_uri method after action is added
    with patch.object(main_window, 'do_show_dataset_details_by_uri',
                      new_callable=MagicMock) as mock_do_show_dataset_details_by_uri:
        # Connect the action
        show_dataset_by_uri_action.connect("activate", main_window.do_show_dataset_details_by_uri)

        # Trigger the action
        show_dataset_by_uri_action.activate(uri_variant)

        # Assert that do_show_dataset_details_by_uri was called once
        mock_do_show_dataset_details_by_uri.assert_called_once_with(show_dataset_by_uri_action, uri_variant)


@pytest.mark.asyncio
async def test_do_show_dataset_details_by_row_index_direct_call(populated_app_with_mock_data, mock_get_datasets):
    """
    Test the do_show_dataset_details_by_row_index method for directly showing dataset details by row index.
    It verifies if the dataset list is correctly populated and the specified dataset details are displayed.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)  # Yield control to allow other async tasks to run
        return False  # Timeout reached if datasets are not loaded within the specified time

    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action and wait for datasets to load
    main_window.activate_action('refresh-view')
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Create a GLib.Variant with a test row index (e.g., 0 for the first row)
    row_index = 1
    row_index_variant = GLib.Variant.new_uint32(row_index)

    # Create and add the show dataset action to the main window of the application
    show_dataset_action = Gio.SimpleAction.new("show-dataset", row_index_variant.get_type())
    main_window.add_action(show_dataset_action)

    # Connect the show dataset action to the do_show_dataset_details_by_row_index method
    show_dataset_action.connect("activate", main_window.do_show_dataset_details_by_row_index)

    # Trigger the show dataset action with the test row index
    show_dataset_action.activate(row_index_variant)

    # Optionally, wait for the UI to update and dataset details to be displayed
    await asyncio.sleep(0.1)  # Adjust this sleep duration as needed

    # Perform assertions to verify that the dataset details are correctly displayed
    selected_dataset = main_window.dataset_list_box.get_row_at_index(row_index).dataset

    expected_uri = mock_get_datasets[1]['uri']
    expected_uuid = mock_get_datasets[1]['uuid']

    assert selected_dataset.uri == expected_uri, f"Expected URI {expected_uri}, got {selected_dataset.uri}"
    assert selected_dataset.uuid == expected_uuid, f"Expected UUID {expected_uuid}, got {selected_dataset.uuid}"

    # Additional assertions can be made here based on the expected behavior and UI element updates

    # Await completion of any remaining asynchronous tasks related to showing the dataset details
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)


@pytest.mark.asyncio
async def test_do_show_dataset_details_by_row_index_action_trigger(running_app):
    """Test if 'show-dataset' action triggers do_show_dataset_details_by_row_index method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Mock dependencies
    mock_dataset_list_box = MagicMock()
    main_window.dataset_list_box = mock_dataset_list_box

    # Setup necessary mocks for the action trigger
    main_window._show_dataset_details_by_row_index = MagicMock()

    # Create and add the action
    row_index_variant = GLib.Variant.new_uint32(1)  # Example index
    show_dataset_action = Gio.SimpleAction.new("show-dataset", row_index_variant.get_type())
    main_window.add_action(show_dataset_action)

    # Patch do_show_dataset_details_by_row_index method after action is added
    with patch.object(main_window, 'do_show_dataset_details_by_row_index',
                      new_callable=MagicMock) as mock_do_show_dataset_details_by_row_index:
        # Connect the action
        show_dataset_action.connect("activate", main_window.do_show_dataset_details_by_row_index)

        # Trigger the action
        show_dataset_action.activate(row_index_variant)

        # Assert that do_show_dataset_details_by_row_index was called once
        mock_do_show_dataset_details_by_row_index.assert_called_once_with(show_dataset_action, row_index_variant)


@pytest.mark.asyncio
async def test_do_search_select_and_show_direct_call(populated_app_with_mock_data, mock_get_datasets):
    """
    Test the do_search_select_and_show method for processing a search directly. It verifies
    if the dataset list is correctly populated, the first dataset is selected, and its
    content is displayed.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)  # Yield control to allow other async tasks to run
        return False  # Timeout reached

    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Create a GLib.Variant with the test search query
    mock_variant = GLib.Variant.new_string("test_search_query")

    # Call the method with the test search query
    main_window.do_search_select_and_show(None, mock_variant)

    # Assertions to check the state of the application after the search
    assert len(main_window.dataset_list_box.get_children()) > 0, "No datasets found in the list box"
    first_dataset_row = main_window.dataset_list_box.get_children()[0]
    dataset = first_dataset_row.dataset

    # Assert that the dataset's URI matches the expected URI
    assert dataset.uri == mock_get_datasets[0]['uri'], f"Expected URI {mock_get_datasets[0]['uri']}, got {dataset.uri}"

    # Await completion of all tasks related to the test
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)


@pytest.mark.asyncio
async def test_do_search_select_and_show_action_trigger(running_app):
    """Test if 'search-select-show' action triggers do_search_select_and_show method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Mock dependencies
    # mock_base_uri_list_box = MagicMock()
    # main_window.base_uri_list_box = mock_base_uri_list_box

    # Setup necessary mocks for the action trigger
    main_window._search_select_and_show = MagicMock()

    # Create and add the action
    search_variant = GLib.Variant.new_string("test_search_query")
    search_select_show_action = Gio.SimpleAction.new("search-select-show", search_variant.get_type())
    main_window.add_action(search_select_show_action)

    # Patch do_search_select_and_show method after action is added
    with patch.object(main_window, 'do_search_select_and_show',
                      new_callable=MagicMock) as mock_do_search_select_and_show:
        # Connect the action
        search_select_show_action.connect("activate", main_window.do_search_select_and_show)

        # Trigger the action
        search_select_show_action.activate(search_variant)

        # Assert that do_search_select_and_show was called once
        mock_do_search_select_and_show.assert_called_once_with(search_select_show_action, search_variant)


@pytest.mark.asyncio
# @pytest.mark.skip(reason="no way of currently testing this")
async def test_do_get_item_direct_call(populated_app_with_local_dataset_data, local_dataset_uri, tmp_path):
    """
    Test the do_get_item method for copying a selected item directly.
    It verifies if the selected item is correctly copied to the specified destination.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)  # Yield control to allow other async tasks to run
        return False  # Timeout reached

    windows = populated_app_with_local_dataset_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action to load datasets
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Select the first dataset
    # first_dataset_row = app_with_mock_dtool_lookup_api_calls_on_local_dataset.main_window.dataset_list_box.get_children()[0]
    # dataset = first_dataset_row.dataset

    # Assuming the first dataset's UUID as the item UUID (Replace this with the actual item UUID logic)

    def get_item_uuid(dataset, relpath):
        for identifier, properties in dataset.generate_manifest()["items"].items():
            if properties["relpath"] == relpath:
                return identifier
        return None

    dataset = dtoolcore.DataSet.from_uri(local_dataset_uri)
    print(dataset.generate_manifest())

    item_uuid = get_item_uuid(dataset, "tiny.png")
    main_window._get_selected_items = lambda: [('item_name', item_uuid)]

    # TODO: assert file size and file content agree with "tiny.png" in data folder

    # Define the destination file path
    dest_file = tmp_path / "destination_file"

    # Create a GLib.Variant with the destination file path
    dest_file_variant = GLib.Variant.new_string(str(dest_file))

    dest_file = tmp_path / "destination_file"
    print(f"Destination file path: {dest_file}")

    # Call the do_get_item method without 'await'
    try:
        main_window.do_get_item(None, dest_file_variant)
    except Exception as e:
        print(f"Error during 'do_get_item': {e}")
        raise

    # Wait for all asynchronous tasks to complete
    await asyncio.sleep(1)  # Increase sleep time if necessary
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        completed_tasks = await asyncio.gather(*pending_tasks, return_exceptions=True)
        for task in completed_tasks:
            if isinstance(task, Exception):
                print(f"Exception in async task: {task}")

    # Verify that the file was copied
    file_exists = dest_file.exists()
    print(f"File exists: {file_exists}")
    assert file_exists, "The item was not copied to the specified destination"

    # Assert File Size
    original_file = Path('test/data/tiny.png')  # Path to the original file
    assert dest_file.stat().st_size == original_file.stat().st_size, "Downloaded file size does not match the original file size"

    # Assert File Content
    with open(dest_file, 'rb') as f_downloaded, open(original_file, 'rb') as f_original:
        assert f_downloaded.read() == f_original.read(), "Downloaded file content does not match the original file content"

    # Assert Modification Time
    current_time = time.time()
    assert dest_file.stat().st_mtime <= current_time, "Downloaded file modification time is in the future"

    # Optional: Assert File Permissions

@pytest.mark.asyncio
async def test_do_get_item_action_trigger(running_app):
    """Test if 'get-item' action triggers do_get_item method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Mock dependencies
    mock_dataset_list_box = MagicMock()
    main_window.dataset_list_box = mock_dataset_list_box
    mock_settings = MagicMock()
    main_window.settings = mock_settings

    # Setup necessary mocks for the action trigger
    mock_dataset = MagicMock()
    mock_dataset.get_item = AsyncMock()
    mock_dataset_list_box.get_selected_row.return_value = MagicMock(dataset=mock_dataset)
    main_window._get_selected_items = MagicMock(return_value=[('item_name', 'item_uuid')])

    # Create and add the action
    dest_file_variant = GLib.Variant.new_string("dummy_path")
    get_item_action = Gio.SimpleAction.new("get-item", dest_file_variant.get_type())
    main_window.add_action(get_item_action)

    # Patch do_get_item method after action is added
    with patch.object(main_window, 'do_get_item', new_callable=MagicMock) as mock_do_get_item:
        # Connect the action
        get_item_action.connect("activate", main_window.do_get_item)

        # Trigger the action
        get_item_action.activate(dest_file_variant)

        # Assert that do_get_item was called once
        mock_do_get_item.assert_called_once_with(get_item_action, dest_file_variant)


@pytest.mark.asyncio
async def test_do_refresh_view_direct_call(running_app):
    """Test the direct call of the do_refresh_view method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Directly call the do_refresh_view method
    main_window.do_refresh_view(None, None)


@pytest.mark.asyncio
async def test_refresh_method_triggered_by_action(running_app):
    """Test if the 'refresh-view' action triggers the refresh method."""

    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Patch the main window's refresh method
    with patch.object(main_window, 'refresh', new_callable=MagicMock) as mock_refresh:
        # Trigger the 'refresh-view' action
        main_window.activate_action('refresh-view')

        # Assert that the refresh method was called once
        mock_refresh.assert_called_once()

# tests for the do_put_annotation action and put-tag action

@pytest.mark.asyncio
async def test_do_put_annotation_direct_call(populated_app_with_local_dataset_data):
    """
    Test the do_put_annotation method for adding an annotation.
    It verifies if the annotation is correctly added to the selected dataset.
    """
    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)
        return False  # Timeout reached

    windows = populated_app_with_local_dataset_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action to load datasets
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Select the first dataset
    main_window.activate_action('select-dataset', GLib.Variant.new_uint32(0))

    # Define annotation key and value
    annotation_key = "test_annotation_key"
    annotation_value = "test_annotation_value"
    annotation_variant = GLib.Variant.new_tuple(GLib.Variant.new_string(annotation_key), GLib.Variant.new_string(annotation_value))

    # Call the do_put_annotation method
    main_window.do_put_annotation(None, annotation_variant)

    # Retrieve the selected dataset to verify the annotation
    selected_row = main_window.dataset_list_box.get_selected_row()
    assert selected_row is not None, "No dataset is selected"
    selected_dataset = selected_row.dataset

     # Retrieve all annotations
    annotations = await selected_dataset.get_annotations()

    # Verify that the new annotation is present
    assert (annotation_key, annotation_value) in annotations.items(), (
        f"Expected annotation ({annotation_key}: {annotation_value}) not found in the dataset"
    )



@pytest.mark.asyncio
async def test_do_put_annotation_action_trigger(running_app):
    """Test if the 'put-annotation' action triggers the do_put_annotation method."""
    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    with patch.object(main_window, '_put_annotation', new_callable=MagicMock) as mock_put_annotation:
        action = main_window.lookup_action("put-annotation")
        action.activate(GLib.Variant.new_tuple(GLib.Variant.new_string("key"), GLib.Variant.new_string("value")))
        mock_put_annotation.assert_called_once_with("key", "value")
        

@pytest.mark.asyncio
async def test_do_put_tag_direct_call(populated_app_with_local_dataset_data):
    """
    Test the do_put_tag method for adding a tag.
    It verifies if the tag is correctly added to the selected dataset.
    """
    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = time.time()
        while time.time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True  # Datasets are loaded
            await asyncio.sleep(0.1)
        return False  # Timeout reached

    windows = populated_app_with_local_dataset_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action to load datasets
    main_window.activate_action('refresh-view')

    # Wait until datasets are loaded
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Select the first dataset
    main_window.activate_action('select-dataset', GLib.Variant.new_uint32(0))

    # Define the tag
    tag_value = "test_tag"
    tag_variant = GLib.Variant.new_string(tag_value)

    # Call the do_put_tag method
    main_window.do_put_tag(None, tag_variant)

    # Retrieve the selected dataset to verify the tag
    selected_row = main_window.dataset_list_box.get_selected_row()
    assert selected_row is not None, "No dataset is selected"
    selected_dataset = selected_row.dataset

    assert tag_value in await selected_dataset.get_tags(), f"Expected tag '{tag_value}' not found in the dataset"

@pytest.mark.asyncio
async def test_do_put_tag_action_trigger(running_app):
    """Test if the 'put-tag' action triggers the do_put_tag method."""
    windows = running_app.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    with patch.object(main_window, '_put_tag', new_callable=MagicMock) as mock_put_tag:
        action = main_window.lookup_action("put-tag")
        action.activate(GLib.Variant.new_string("tag"))
        mock_put_tag.assert_called_once_with("tag")


@pytest.mark.asyncio
async def test_do_select_dataset_row_by_uri_direct_call(populated_app_with_mock_data):
    """
    Test the do_select_dataset_row_by_uri method for directly selecting a dataset row by URI.
    It verifies if the dataset list is correctly populated and the specified dataset row is selected.
    """

    async def wait_for_datasets_to_load(list_box, timeout=10):
        start_time = asyncio.get_event_loop().time()
        while asyncio.get_event_loop().time() - start_time < timeout:
            if len(list_box.get_children()) > 0:
                return True
            await asyncio.sleep(0.1)
        return False

    windows = populated_app_with_mock_data.get_windows()
    main_window = [w for w in windows if isinstance(w, MainWindow)][0]

    # Trigger the 'refresh-view' action and wait for datasets to load
    main_window.activate_action('refresh-view')
    datasets_loaded = await wait_for_datasets_to_load(main_window.dataset_list_box)
    assert datasets_loaded, "Datasets were not loaded in time"

    # Define a URI that corresponds to a dataset in the populated dataset list
    test_uri = "s3://test-bucket/1a1f9fad-8589-413e-9602-5bbd66bfe675"

    # Create a GLib.Variant with the test URI
    uri_variant = GLib.Variant.new_string(test_uri)

    dummy_action = Gio.SimpleAction.new("select-dataset-by-uri", uri_variant.get_type())

    # Call the do_select_dataset_row_by_uri method with the test URI
    main_window.do_select_dataset_row_by_uri(dummy_action , uri_variant)

    # Retrieve the dataset that should be selected based on the URI
    index = main_window.dataset_list_box.get_row_index_from_uri(test_uri)
    selected_row = main_window.dataset_list_box.get_row_at_index(index)
    selected_dataset = selected_row.dataset if selected_row is not None else None

    # Assert that the dataset with the given URI is selected
    assert selected_dataset is not None, "No dataset was selected"
    assert selected_dataset.uri == test_uri, f"Expected URI {test_uri}, got {selected_dataset.uri}"

    # Await completion of all tasks related to the test
    pending_tasks = [task for task in asyncio.all_tasks() if task is not asyncio.current_task()]
    if pending_tasks:
        await asyncio.gather(*pending_tasks)




import subprocess

def get_number_of_tests():
    try:
        result = subprocess.run(
            ["pytest", "--collect-only"],
            capture_output=True,
            text=True,
            check=True,
        )
        for line in result.stdout.splitlines():
            if line.startswith("collected"):
                # Extract the number of tests from the line
                return int(line.split()[1])
    except subprocess.CalledProcessError as e:
        print("Error while running pytest:", e.stderr)
        return None

if __name__ == "__main__":
    num_tests = get_number_of_tests()
    if num_tests is not None:
        print(num_tests)



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging

from math import pi

from gi.repository import GObject, Gdk, Gtk

_log = logging.getLogger(__name__)


class DtoolProgressChart(Gtk.DrawingArea):
    __gtype_name__ = 'DtoolProgressChart'

    _background_color = Gdk.color_parse('lightgray')
    _pie_color = Gdk.color_parse('gray')

    def __init__(self, *args, **kwargs):
        self._fraction = kwargs.pop('fraction', 0)
        super().__init__(*args, **kwargs)
        self.connect('draw', self.on_draw)

    @property
    def fraction(self):
        return self._fraction

    def set_text(self, value):
        _log.info(f'Received text: {value}')

    def set_fraction(self, value):
        self._fraction = value
        self.queue_draw()

    def _cairo_scale(self, area, context):
        w, h = area.get_allocated_width(), area.get_allocated_height()
        s = min(w, h)
        context.translate(w / 2, h / 2)
        context.scale(s, s)

    def on_draw(self, area, context):
        # context.set_source_rgb(1, 1, 1)
        # context.paint()

        # Set scale transformation
        self._cairo_scale(area, context)

        # Draw background circle
        context.set_source_rgb(*self._background_color.to_floats())
        context.arc(0, 0, 0.5, 0, 2 * pi)
        context.close_path()
        context.fill()

        # Draw pie
        context.set_source_rgb(*self._pie_color.to_floats())
        context.arc(0, 0, 0.5, 3 * pi / 2, 3 * pi / 2 + 2 * pi * self._fraction)
        context.line_to(0, 0)
        context.close_path()
        context.fill()


GObject.type_register(DtoolProgressChart)



# Configuration file for the Sphinx documentation builder.

# -- Prepare

import shutil
import os

root = os.path.dirname(__file__)

# Define the source and destination paths for images
data_source_dir = os.path.join(root, '..', '..', 'data')
data_destination_dir = os.path.join(root, 'data')

# Copy images from source to destination
shutil.rmtree(data_destination_dir, ignore_errors=True)  # Remove existing destination directory
shutil.copytree(data_source_dir, data_destination_dir)  # Copy images

# -- Project information

project = 'dtool-lookup-gui'
copyright = '2023, livMatS'
author = 'Antoine Sanner, Ashwin Vazhappilly, Johannes Laurin Hörmann, Lars Pastewka, Michal Rössler, Wolfram Nöhring'


import dtool_lookup_gui
version = dtool_lookup_gui.__version__
release = dtool_lookup_gui.__version__

# release = '0.1'
# version = '0.1.0'

# -- General configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'myst_parser'
]

intersphinx_mapping = {
    'python': ('https://docs.python.org/3/', None),
    'sphinx': ('https://www.sphinx-doc.org/en/master/', None),
}
intersphinx_disabled_domains = ['std']

templates_path = ['_templates']

# -- Options for HTML output

html_theme = 'sphinx_rtd_theme'

# -- Options for EPUB output
epub_show_urls = 'footnote'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['data']



#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import os

from gi.repository import Gtk

from dtoolcore.utils import NAME_VALID_CHARS_LIST


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/dataset_name_dialog.ui')
class DatasetNameDialog(Gtk.Window):
    __gtype_name__ = 'DtoolDatasetNameDialog'

    name_label = Gtk.Template.Child()
    name_entry = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        self._on_confirmation = kwargs.pop('on_confirmation')
        super().__init__(*args, **kwargs)
        valid_chars = ' '.join(NAME_VALID_CHARS_LIST)
        self.name_label.set_text(f'Please provide a name for the new dataset. The name must be unique to the local '
                                 f'directory where the dataset is created by must not be unique across a cloud '
                                 f'storage system. The name may only contain the following characters: {valid_chars}')

    @Gtk.Template.Callback()
    def on_apply_clicked(self, widget):
        self._on_confirmation(self.name_entry.get_text())
        self.destroy()

    @Gtk.Template.Callback()
    def on_cancel_clicked(self, widget):
        self.destroy()


#
# Copyright 2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

from gi.repository import Gtk, Pango


class DtoolBaseURIRow(Gtk.ListBoxRow):
    __gtype_name__ = 'DtoolBaseURIRow'

    _margin = 3

    _folder_symbols = {
        'file': 'folder-symbolic',
    }

    def __init__(self, base_uri, *args, **kwargs):
        self._base_uri = base_uri
        self._task = None

        on_activate = kwargs.pop('on_activate', None)
        on_configure = kwargs.pop('on_configure', None)
        on_remove = kwargs.pop('on_remove', None)

        super().__init__(*args, **kwargs)

        hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, margin_top=self._margin, margin_bottom=self._margin,
                       margin_start=self._margin, margin_end=self._margin)
        image = Gtk.Image.new_from_icon_name(self._folder_symbols[base_uri.scheme]
                                             if base_uri.scheme in self._folder_symbols
                                             else 'network-server-symbolic', Gtk.IconSize.BUTTON)
        image.set_padding(12, 12)
        hbox.pack_start(image, False, False, 0)
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        label = Gtk.Label(xalign=0)
        if base_uri.scheme == 'file':
            label.set_markup(f'<b>{base_uri.uri_name}</b>')
        else:
            label.set_markup(f'<b>{base_uri}</b>')
        label.set_ellipsize(Pango.EllipsizeMode.END)
        vbox.pack_start(label, True, True, 0)
        self._info_label = Gtk.Label(xalign=0)
        self._info_label.set_text('---')
        self._info_label.set_ellipsize(Pango.EllipsizeMode.END)
        # if info['local'] and info['remote']:
        #    label.set_markup('This endpoint is reported by the lookup server and it is configured locally.')
        # elif not info['local'] and info['remote']:
        #    label.set_markup('This endpoint is reported by the lookup server, but it is not configured locally.')
        # else:
        #    label.set_markup('This endpoint is configured locally.')
        vbox.pack_start(self._info_label, True, True, 0)
        hbox.pack_start(vbox, True, True, 0)
        self._spinner = Gtk.Spinner(margin_top=self._margin, margin_bottom=self._margin,
                                    margin_start=self._margin, margin_end=self._margin)
        hbox.pack_start(self._spinner, False, False, 0)
        if on_configure is not None:
            button = Gtk.Button(image=Gtk.Image.new_from_icon_name('emblem-system-symbolic', Gtk.IconSize.BUTTON))
            button.connect('clicked', on_configure)
            # We currently can only configure S3 endpoints through the GUI
            # button.set_sensitive(parsed_uri.scheme == 's3')
            # button.base_uri = parsed_uri
            hbox.pack_end(button, False, False, 0)
        if on_remove is not None:
            button = Gtk.Button(image=Gtk.Image.new_from_icon_name('window-close-symbolic', Gtk.IconSize.BUTTON))
            button.connect('clicked', on_remove)
            hbox.pack_end(button, False, False, 0)
        self.add(hbox)

        if on_activate is not None:
            self.connect('activate', on_activate)

    @property
    def base_uri(self):
        return self._base_uri

    @property
    def info_label(self):
        return self._info_label

    def start_spinner(self):
        self._spinner.start()

    def stop_spinner(self):
        self._spinner.stop()

    @property
    def task(self):
        return self._task

    @task.setter
    def task(self, task):
        self._taks = task


#
# Copyright 2021 Lars Pastewka
#           2021 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

"""Progressbar"""

import logging

from gi.repository import Gtk

from contextlib import AbstractContextManager
from typing import Optional

logger = logging.getLogger(__name__)


class ProgressBar(AbstractContextManager):
    """Mimics click.progressbar". Just log messages if no progressbar specified."""
    def __init__(self, length=None, label=None,
                 pb: Optional[Gtk.ProgressBar] = None):
        # if pb is None:
        #    pb = Gtk.ProgressBar(show_text=True, text=None)
        self._pb = pb
        self._item_show_func = None
        self._label_template = '{label:}'
        self._label_item_template = '{label:} ({item:})'
        self._label = label
        self._length = length
        self._step = 0

    def __enter__(self):
        if self._pb is not None:
            if hasattr(self._pb, 'set_fraction'):
                self._pb.set_fraction(0.0)
            if hasattr(self._pb, 'set_step'):
                self._pb.set_step(0, self._length)
        logger.info(f"Progress fraction 0.0")
        self._set_text()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, label):
        self._label = label
        self._set_text()

    @property
    def item_show_func(self):
        return self._item_show_func

    @item_show_func.setter
    def item_show_func(self, item_show_func):
        self._item_show_func = item_show_func
        self._set_text()

    def update(self, step):
        self._step += step
        fraction = float(self._step) / float(self._length)
        if self._pb is not None:
            if hasattr(self._pb, 'set_fraction'):
                self._pb.set_fraction(fraction)
            if hasattr(self._pb, 'set_step'):
                self._pb.set_step(self._step, self._length)
        logger.info(f"Progress fraction {fraction}")
        self._set_text()

    def _set_text(self):
        if self._label is not None and self._item_show_func is not None:
            text = self._label_item_template.format(
                label=self._label, item=self.item_show_func(self._step))
            if self._pb is not None:
                self._pb.set_text(text)
        if self._label is not None:
            text = self._label_template.format(label=self._label)
            if self._pb is not None:
                self._pb.set_text(text)
        else:
            text = None
            if self._pb is not None:
                self._pb.set_show_text(False)

        if text is not None:
            logger.info(text)


#
# Copyright 2021-2022 Johannes Laurin Hörmann
#           2021 Lars Pastewka
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#

import logging
import os

from gi.repository import GLib, Gdk, Gtk, GtkSource

from ..utils.query import (
    single_line_sanitize_query_text,
    multi_line_sanitize_query_text)

logger = logging.getLogger(__name__)


@Gtk.Template(filename=f'{os.path.dirname(__file__)}/search_popover.ui')
class DtoolSearchPopover(Gtk.Popover):
    __gtype_name__ = 'DtoolSearchPopover'

    search_text_view = Gtk.Template.Child()
    search_button = Gtk.Template.Child()
    cancel_button = Gtk.Template.Child()

    def __init__(self, *args, search_entry : Gtk.SearchEntry, **kwargs):
        # on_show_clicked = kwargs.pop('on_show_clicked', None)
        super().__init__(*args, **kwargs)
        # if on_show_clicked is not None:
            # self.show_dataset_button.connect('clicked', on_show_clicked)

        self.search_entry = search_entry
        self.search_entry_buffer = self.search_entry.get_buffer()
        self.search_text_buffer = self.search_text_view.get_buffer()

        # set up some pseudo highlighting for the log window
        lang_manager = GtkSource.LanguageManager()
        self.search_text_buffer.set_language(lang_manager.get_language("json"))
        self.search_text_buffer.set_highlight_syntax(True)
        self.search_text_buffer.set_highlight_matching_brackets(True)

        self.set_relative_to(self.search_entry)

    def popup_at(self, widget):
        search_text = self.search_entry_buffer.get_text()
        try:
            sanitized_search_text = multi_line_sanitize_query_text(search_text)
        except:
            logger.debug("Not a valid query.")
            sanitized_search_text = search_text
        self.search_text_buffer.set_text(sanitized_search_text, -1)

        self.set_relative_to(widget)
        logger.debug("Show search popover.")
        self.popup()

    @Gtk.Template.Callback()
    def on_search_button_clicked(self, button):
        """"Update search bar text when clicking search button in search popover."""
        start_iter = self.search_text_buffer.get_start_iter()
        end_iter = self.search_text_buffer.get_end_iter()
        search_text = self.search_text_buffer.get_text(start_iter, end_iter, True)
        try:
            sanitized_search_text = single_line_sanitize_query_text(search_text)
        except:
            logger.debug("Not a valid query.")
            sanitized_search_text = search_text
        self.search_entry_buffer.set_text(sanitized_search_text, -1)
        self.popdown()
        self.get_action_group("win").activate_action('search-select-show', GLib.Variant.new_string(sanitized_search_text))

    @Gtk.Template.Callback()
    def on_cancel_button_clicked(self, button):
        self.popdown()



#
# Copyright 2022-2023 Johannes Laurin Hörmann
#
# ### MIT license
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import asyncio
import json
import logging
import os
import pytest
import pytest_asyncio
import uuid

import threading
import warnings

import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GtkSource', '4')
from gi.repository import GLib, GObject, Gio, Gtk, GtkSource, GdkPixbuf

from gi.events import GLibEventLoopPolicy, GLibEventLoop


from dtool_lookup_gui.main import Application

from unittest.mock import patch

# logging.basicConfig(level=logging.DEBUG)

logger = logging.getLogger(__name__)


_HERE = os.path.dirname(os.path.abspath(__file__))

# dtool_lookup_api Config

DSERVER_URL = "https://localhost:5000/lookup"
DSERVER_TOKEN = r"eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJmcmVzaCI6ZmFsc2UsImlhdCI6MTcwMTY4OTU5MywianRpIjoiYTdkN2Y5ZWItZGI3MS00YjExLWFhMzktZGQ2YzgzOTJmOWE4IiwidHlwZSI6ImFjY2VzcyIsInN1YiI6InRlc3R1c2VyIiwibmJmIjoxNzAxNjg5NTkzLCJleHAiOjE3MDE2OTMxOTN9.t9SQ00ecZRc-pspz-Du7321xfWIzgTTFKobkNed1CuQYHvtNrc3vdHYbqCWaYCqZpEVF8RlltldT4Lookx6vNgnW4olpiS2KTZ-X2asMhn7SShDtUJuU54CGeViWzYX_V_Pzckoe_cgjFkOutRvnwy_072Whnmc0TwYojwNqUScAIJRu0pzym84JngloXfdI7r25GcRVNtzsGUl7DDfrIz4aSOeVDAVEXhPjgEatKsvNdVZl1DIJsTZpuI7Jh7ZW1WsyjonqHR0J0kIVQn9imQyLyS9_CtmURBQ3kabx6cxhpx5LADrzLutSu24eA4FyECOdzjJ3SPGb9nIVTEDxQg"
DSERVER_TOKEN_GENERATOR_URL = "https://localhost:5000/token"
DSERVER_USERNAME = "testuser"
DSERVER_PASSWORD = "test_password"
DSERVER_VERIFY_SSL = True

AFFIRMATIVE_EXPRESSIONS = ['true', '1', 'y', 'yes', 'on']
NEGATIVE_EXPRESSIONS = ['false', '0', 'n', 'no', 'off']

logger = logging.getLogger(__name__)


class CustomGLibEventLoopPolicy(GLibEventLoopPolicy):
    def set_event_loop(self, loop):
        """Set the event loop for the current context (python thread) to loop.

        This is only permitted if the thread has no thread default main context
        with the main thread using the default main context.
        """

        # Only accept glib event loops, otherwise things will just mess up
        assert loop is None or isinstance(loop, GLibEventLoop)

        ctx = ctx_td = GLib.MainContext.get_thread_default()
        if ctx is None and threading.current_thread() is threading.main_thread():
            ctx = GLib.MainContext.default()

        if loop is None:
            # We do permit unsetting the current loop/context
            old = self._loops.pop(hash(ctx), None)
            if old:
                if hash(old._context) != hash(ctx):
                    warnings.warn('GMainContext was changed unknowingly by asyncio integration!', RuntimeWarning)
                if ctx_td:
                    GLib.MainContext.pop_thread_default(ctx_td)
        else:
            # # Only allow attaching if the thread has no main context yet
            # if ctx:
            #     raise RuntimeError(
            #         'Thread %r already has a main context, get_event_loop() will create a new loop if needed'
            #         % threading.current_thread().name)

            GLib.MainContext.push_thread_default(loop._context)
            self._loops[hash(loop._context)] = loop


class MockDtoolLookupAPIConfig():
    """Mock dtool configuration without touching the local config file."""

    def __init__(self, *args, **kwargs):
        self._lookup_server_url = DSERVER_URL
        self._lookup_server_token = DSERVER_URL
        self._lookup_server_token_generator_url = DSERVER_TOKEN_GENERATOR_URL
        self._lookup_server_username = DSERVER_USERNAME
        self._lookup_server_password = DSERVER_PASSWORD
        self._lookup_server_verify_ssl = DSERVER_VERIFY_SSL

    @property
    def lookup_url(self):
        return self._lookup_server_url

    @lookup_url.setter
    def lookup_url(self, value):
        self._lookup_server_url = value

    # optional
    @property
    def token(self):
        return self._lookup_server_token

    @token.setter
    def token(self, token):
        self._lookup_server_token = token

    @property
    def auth_url(self):
        return self._lookup_server_token_generator_url

    @auth_url.setter
    def auth_url(self, value):
        self._lookup_server_token_generator_url = value

    @property
    def username(self):
        return self._lookup_server_username

    @username.setter
    def username(self, value):
        self._lookup_server_username = value

    @property
    def password(self):
        return self._lookup_server_password

    @password.setter
    def password(self, value):
        self._lookup_server_password = value

    @property
    def verify_ssl(self):
        return self._lookup_server_verify_ssl

    @verify_ssl.setter
    def verify_ssl(self, value):
        self._lookup_server_verify_ssl = value


@pytest.fixture(scope="function")
def mock_token(scope="function"):
    """Provide a mock authentication token"""
    yield "eyJ0eXAiOiJKV1QiLCJhbGciOiJSUzI1NiJ9.eyJmcmVzaCI6ZmFsc2UsImlhdCI6MTcwMTY4OTU5MywianRpIjoiYTdkN2Y5ZWItZGI3MS00YjExLWFhMzktZGQ2YzgzOTJmOWE4IiwidHlwZSI6ImFjY2VzcyIsInN1YiI6InRlc3R1c2VyIiwibmJmIjoxNzAxNjg5NTkzLCJleHAiOjE3MDE2OTMxOTN9.t9SQ00ecZRc-pspz-Du7321xfWIzgTTFKobkNed1CuQYHvtNrc3vdHYbqCWaYCqZpEVF8RlltldT4Lookx6vNgnW4olpiS2KTZ-X2asMhn7SShDtUJuU54CGeViWzYX_V_Pzckoe_cgjFkOutRvnwy_072Whnmc0TwYojwNqUScAIJRu0pzym84JngloXfdI7r25GcRVNtzsGUl7DDfrIz4aSOeVDAVEXhPjgEatKsvNdVZl1DIJsTZpuI7Jh7ZW1WsyjonqHR0J0kIVQn9imQyLyS9_CtmURBQ3kabx6cxhpx5LADrzLutSu24eA4FyECOdzjJ3SPGb9nIVTEDxQg"


@pytest.fixture(scope="function")
def mock_get_datasets():
    """Provide a mock dataset list"""
    with open(os.path.join(_HERE, 'data', 'mock_get_datasets_response.json'), 'r') as f:
        dataset_list = json.load(f)

    yield dataset_list


@pytest.fixture(scope="function")
def mock_get_manifest():
    """Provide a mock manifest"""
    with open(os.path.join(_HERE, 'data', 'mock_get_manifest_response.json'), 'r') as f:
        manifest = json.load(f)

    yield manifest


@pytest.fixture(scope="function")
def mock_get_readme():
    """Provide mock readme content"""
    with open(os.path.join(_HERE, 'data', 'mock_get_readme_response.json'), 'r') as f:
        readme = json.load(f)

    yield readme


@pytest.fixture(scope="function")
def mock_get_tags():
    """Provide mock tags"""
    yield ["tag1", "tag2"]


@pytest.fixture(scope="function")
def mock_get_annotations():
    """Provide mock annotations"""
    yield {"annotation1": "value1", "annotation2": "value2"}


@pytest.fixture(scope="function")
def mock_get_config():
    """Provide a mock server config info"""
    with open(os.path.join(_HERE, 'data', 'mock_get_config_response.json'), 'r') as f:
        config_info = json.load(f)

    yield config_info


@pytest.fixture(scope="function")
def mock_get_versions():
    """Provide a mock server versions info"""
    with open(os.path.join(_HERE, 'data', 'mock_get_versions_response.json'), 'r') as f:
        config_info = json.load(f)

    yield config_info




@pytest.fixture(scope="function")
def local_dataset_uri(tmp_path):
    """Create a proper local dataset to test against."""

    from dtoolcore import ProtoDataSet, generate_admin_metadata
    from dtoolcore.storagebroker import DiskStorageBroker

    name = "test_dataset"
    admin_metadata = generate_admin_metadata(name)
    dest_uri = DiskStorageBroker.generate_uri(
        name=name,
        uuid=admin_metadata["uuid"],
        base_uri=str(tmp_path))

    # create a proto dataset
    proto_dataset = ProtoDataSet(
        uri=dest_uri,
        admin_metadata=admin_metadata,
        config_path=None)
    proto_dataset.create()
    proto_dataset.put_readme("")

    # put two local files
    handle = "sample_dataset_content.txt"
    local_file_path = os.path.join(_HERE, 'data', handle)
    proto_dataset.put_item(local_file_path, handle)

    second_fname = "tiny.png"
    local_file_path = os.path.join(_HERE, 'data', second_fname)
    proto_dataset.put_item(local_file_path, second_fname)

    proto_dataset.freeze()

    yield dest_uri

@pytest.fixture(scope="function")
def event_loop_policy():
    policy = CustomGLibEventLoopPolicy()
    logger.debug("Set asyncio event loop policy to %s.", policy)
    asyncio.set_event_loop_policy(policy)
    yield
    asyncio.set_event_loop_policy(None)


@pytest_asyncio.fixture(loop_scope="function", scope="function")
async def app():
    logger.debug("Register GtkSource.View.")
    GObject.type_register(GtkSource.View)

    event_loop = asyncio.get_running_loop()
    event_loop.set_debug(True)

    logger.debug("Create app within %s.", event_loop)
    app = Application(loop=event_loop,
                      flags=(Gio.ApplicationFlags.NON_UNIQUE | Gio.ApplicationFlags.HANDLES_COMMAND_LINE))

    yield app


@pytest_asyncio.fixture(loop_scope="function", scope="function")
async def running_app(app):
    logger.debug("Register Gtk application.")
    app.register()
    logger.debug("Called app.register().")

    await app.wait_for_startup()
    logger.debug("App finished startup.")

    app.activate()
    logger.debug("Called app.activate().")

    await app.wait_for_activation()
    logger.debug("App finished activation.")

    yield app
    logger.debug("Test finished.")

    logger.debug("Wait for 3 seconds.")
    await asyncio.sleep(3)

    logger.debug("Shutdown.")
    await app.shutdown()

    logger.debug("Wait for app to finish shutdown.")
    await app.wait_for_shutdown()


@pytest_asyncio.fixture(loop_scope="function", scope="function")
async def populated_app_with_mock_data(
        app, mock_get_datasets,
        mock_get_manifest, mock_get_readme, mock_get_tags, mock_get_annotations, mock_get_config):
    """Replaces lookup api calls with mock methods that return fake lists of datasets."""

    import dtool_lookup_api.core.config
    dtool_lookup_api.core.config.Config = MockDtoolLookupAPIConfig()

    # TODO: figure out whether mocked methods work as they should
    with (
            patch("dtool_lookup_api.core.LookupClient.CredentialsBasedLookupClient.authenticate",
                  eturn_value=mock_token),
            patch("dtool_lookup_api.core.LookupClient.ConfigurationBasedLookupClient.connect",
                  return_value=None),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_datasets",
                  return_value=mock_get_datasets),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_graph_by_uuid",
                  return_value=[]),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_manifest",
                  return_value=mock_get_manifest),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_readme",
                  return_value=mock_get_readme),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_tags",
                  return_value=mock_get_tags),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_annotations",
                  return_value=mock_get_annotations),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_config",
                  return_value=mock_get_config),
            patch("dtool_lookup_api.core.LookupClient.ConfigurationBasedLookupClient.has_valid_token",
                  return_value=True)
        ):
        logger.debug("Register Gtk application.")
        app.register()
        logger.debug("Called app.register().")

        await app.wait_for_startup()
        logger.debug("App finished startup.")

        app.activate()
        logger.debug("Called app.activate().")

        await app.wait_for_activation()
        logger.debug("App finished activation.")

        yield app
        logger.debug("Test finished.")

        logger.debug("Wait for 3 seconds.")
        await asyncio.sleep(3)

        logger.debug("Shutdown.")
        await app.shutdown()

        logger.debug("Wait for app to finish shutdown.")
        await app.wait_for_shutdown()


@pytest_asyncio.fixture(loop_scope="function", scope="function")
async def populated_app_with_local_dataset_data(
        app, local_dataset_uri,
        mock_token, mock_get_readme, mock_get_tags, mock_get_annotations, mock_get_config):
    """Replaces lookup api calls with mock methods that return fake lists of datasets."""

    import dtool_lookup_api.core.config
    dtool_lookup_api.core.config.Config = MockDtoolLookupAPIConfig()

    from dtoolcore import DataSet
    dataset = DataSet.from_uri(local_dataset_uri)

    dataset_info = dataset._admin_metadata
    # this returns a dict like
    #   {'uuid': 'f9904c40-aff3-43eb-b062-58919c00062a',
    #    'dtoolcore_version': '3.18.2',
    #    'name': 'test_dataset',
    #    'type': 'dataset',
    #    'creator_username': 'jotelha',
    #    'created_at': 1702902057.239987,
    #    'frozen_at': 1702902057.241831
    #   }

    dataset_info["uri"] = dataset.uri
    dataset_info["base_uri"] = os.path.dirname(local_dataset_uri)

    dataset_list = [dataset_info]

    manifest = dataset.generate_manifest()

    with (
            patch("dtool_lookup_api.core.LookupClient.CredentialsBasedLookupClient.authenticate",
                  return_value=mock_token),
            patch("dtool_lookup_api.core.LookupClient.ConfigurationBasedLookupClient.connect",
                  return_value=None),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_datasets",
                  return_value=dataset_list),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_graph_by_uuid",
                  return_value=[]),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_manifest",
                  return_value=manifest),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_readme",
                  return_value=mock_get_readme),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_tags",
                  return_value=mock_get_tags),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_annotations",
                  return_value=mock_get_annotations),
            patch("dtool_lookup_api.core.LookupClient.TokenBasedLookupClient.get_config",
                  return_value=mock_get_config),
            patch("dtool_lookup_api.core.LookupClient.ConfigurationBasedLookupClient.has_valid_token",
                  return_value=True)
        ):
        logger.debug("Register Gtk application.")
        app.register()
        logger.debug("Called app.register().")

        await app.wait_for_startup()
        logger.debug("App finished startup.")

        app.activate()
        logger.debug("Called app.activate().")

        await app.wait_for_activation()
        logger.debug("App finished activation.")

        yield app
        logger.debug("Test finished.")

        logger.debug("Wait for 3 seconds.")
        await asyncio.sleep(3)

        logger.debug("Shutdown.")
        await app.shutdown()

        logger.debug("Wait for app to finish shutdown.")
        await app.wait_for_shutdown()
